#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <cuda_runtime.h>
#include "cublas_v2.h"

#include "../misc/cuda_util.h"

#define IDX2C(i,j,ld) (((j)*(ld))+(i))

int my_cublas_sgemm(int n,
                  const float alpha,
                  float* a_h, int lda,
                  float* b_h, int ldb,
                  const float beta,
                  float* c_h, int ldc,
                  float* elapsedTime)
{

   size_t size = n * n * sizeof(float);

   //allocate device memory
   float *a_d=0, *b_d=0, *c_d=0;
   HANDLE_ERROR( cudaMalloc ((void**)&a_d, size) );
   HANDLE_ERROR( cudaMalloc ((void**)&b_d, size) );
   HANDLE_ERROR( cudaMalloc ((void**)&c_d, size) );

   if (!a_d || !b_d || !c_d) {
      printf ("device memory allocation failed");
      exit(-1);
   }

   //initialize CUBLAS
   cublasHandle_t handle;
   HANDLE_ERROR( cublasCreate(&handle) );

   //create stream
   cudaStream_t stream;
   HANDLE_ERROR( cudaStreamCreate(&stream) );

   //all cublas tasks will be scheduled to this stream
   HANDLE_ERROR( cublasSetStream(handle, stream) );

   //start timing
   cudaEvent_t start,stop;
   cudaEventCreate(&start); cudaEventCreate(&stop);
   cudaEventRecord(start,stream);

   //host to device
   HANDLE_ERROR( cudaMemcpyAsync(a_d, a_h, size, cudaMemcpyHostToDevice, stream) );
   HANDLE_ERROR( cudaMemcpyAsync(b_d, b_h, size, cudaMemcpyHostToDevice, stream) );
   HANDLE_ERROR( cudaMemcpyAsync(c_d, c_h, size, cudaMemcpyHostToDevice, stream) );

   //start computation
   HANDLE_ERROR( cublasSgemm(   handle,
                  CUBLAS_OP_N, CUBLAS_OP_N,
                  n, n, n,
                  &alpha,
                  a_d, lda,
                  b_d, ldb,
                  &beta,
                  c_d, ldc) );

   //device to host
   HANDLE_ERROR( cudaMemcpyAsync(c_h, c_d, size, cudaMemcpyDeviceToHost, stream) );

   //stop timing
   cudaEventRecord(stop,stream);
   cudaEventSynchronize(stop);
   cudaEventElapsedTime(elapsedTime,start,stop);


   //free memory
   cudaFree (a_d); cudaFree (b_d); cudaFree (c_d);

   HANDLE_ERROR( cublasDestroy(handle) );
   HANDLE_ERROR( cudaStreamDestroy(stream) );

   return 0;
}

