#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <mkl.h>
#include <omp.h>

#include <mkl_cblas.h>

#include "cublas_sgemm.h"
#include "../misc/util.h"

#define IDX2C(i,j,ld) (((j)*(ld))+(i))


int main(int argc, char *argv[])
{
   int n = 2048;
   if(argc > 1)
      n = atoi(argv[1]);

   const float alpha = 1.2f; 
   const float beta = 1.1f;
   int lda, ldb, ldc;
   lda = ldb = ldc = n;



   //allocate host memory
   float *a_h=0, *b_h=0, *c_h=0, *c_copy=0;
   unsigned int size = n * n * sizeof (float);
   a_h = (float *)malloc (size);
   b_h = (float *)malloc (size);
   c_h = (float *)malloc (size);
   c_copy = (float *)malloc (size);

   if (!a_h || !b_h || !c_h) {
      printf ("host memory allocation failed");
      exit(-1);
   }

   //initialize matrices
#pragma omp parallel for
   for (int j = 0; j < n; j++) {
      for (int i = 0; i < n; i++) {
         a_h[IDX2C(i,j,n)] = ((float)rand())/RAND_MAX;
         b_h[IDX2C(i,j,n)] = ((float)rand())/RAND_MAX;
         c_h[IDX2C(i,j,n)] = ((float)rand())/RAND_MAX;
         c_copy[IDX2C(i,j,n)] = c_h[IDX2C(i,j,n)];
      }
   }

   int nIter = 10;

   /********************************************
    * call CUBLAS
    ********************************************/
   float et_cublas = FLT_MAX;
   for(int i=0;i < nIter; ++i)
   {
      float et_tmp;
      my_cublas_sgemm(n,
            alpha,
            a_h, lda,
            b_h, ldb,
            beta,
            c_h, ldc,
            &et_tmp);

      //take minimum over nIter runs
      if(et_tmp < et_cublas)
         et_cublas = et_tmp;
   }

   /********************************************
    * call MKL
    ********************************************/
   double et_mkl = DBL_MAX;
   for(int i=0;i < nIter; ++i)
   {
      double start_mkl = dsecnd();
      cblas_sgemm(  CblasColMajor, CblasNoTrans, CblasNoTrans, 
            n, n, n, 
            alpha, 
            a_h, lda, 
            b_h, ldb,
            beta, 
            c_copy, ldc);
      double end_mkl = dsecnd();

      float et_tmp = (end_mkl - start_mkl) * 1000.0f;
      //take minimum over nIter runs
      if(et_tmp < et_mkl)
         et_mkl = et_tmp;
   }

   //display performance results
   double flopsPerMatrixMul = 2.0 * (double)n * (double)n * (double)n;
   double gigaFlops_cublas = (flopsPerMatrixMul * 1.0e-9f) / (et_cublas / 1000.0f);
   double gigaFlops_mkl = (flopsPerMatrixMul * 1.0e-9f) / (et_mkl / 1000.0f);
   printf("CUBLAS computation took %f ms and achieved %f GFLOPS.\n", et_cublas, gigaFlops_cublas);
   printf("MKL computation took %f ms and achieved %f GFLOPS.\n", et_mkl, gigaFlops_mkl);
   printf("Speedup (MKL/CUBLAS): %f\n", et_mkl / et_cublas);

   printf("%d %f %f\n", n, gigaFlops_cublas, gigaFlops_mkl);

   //check the result
   float max_error = 0.0f;
   float max = 1.0f;
   for(int j=0;j < n; ++j){
      for(int i=0;i < n; ++i){
         float error = fabs(c_copy[IDX2C(i,j,n)] - c_h[IDX2C(i,j,n)]);
         if (error > max_error){
            max_error = error;
            max = c_copy[IDX2C(i,j,n)];
         }
      }
   }
   if(fabs(max_error/max) < 1e-5)
      printf_green("RESULT OKAY!\n");
   else
      printf_red("RESULT FAILED!\n");


   //free memory
   free(a_h); free(b_h); free(c_h); free(c_copy);

   return 0;
}

